package betterwithmods.common.blocks.mechanical.tile;

import betterwithmods.api.block.ISoulSensitive;
import betterwithmods.api.capabilities.CapabilityMechanicalPower;
import betterwithmods.api.tile.IMechanicalPower;
import betterwithmods.client.model.filters.ModelWithResource;
import betterwithmods.client.model.render.RenderUtils;
import betterwithmods.common.blocks.tile.IMechSubtype;
import betterwithmods.common.blocks.tile.SimpleStackHandler;
import betterwithmods.common.blocks.tile.TileEntityVisibleInventory;
import betterwithmods.common.registry.HopperFilters;
import betterwithmods.common.registry.HopperInteractions;
import betterwithmods.util.InvUtils;
import betterwithmods.util.MechanicalUtil;
import betterwithmods.util.WorldUtils;
import net.minecraft.block.Block;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.item.EntityXPOrb;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.SoundEvents;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.EntitySelectors;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.items.IItemHandler;

import javax.annotation.Nonnull;
import java.util.List;
import java.util.Optional;

public class TileEntityFilteredHopper extends TileEntityVisibleInventory implements IMechSubtype, IMechanicalPower {

    private final int STACK_SIZE = 8;
    public int filterType;
    public int soulsRetained;
    private int ejectCounter, ejectXPCounter;
    private int experienceCount, maxExperienceCount = 1000;
    public byte power;

    public TileEntityFilteredHopper() {
        this.ejectCounter = 0;
        this.experienceCount = 0;
        this.ejectXPCounter = 10;
        this.filterType = 0;
        this.soulsRetained = 0;
        this.occupiedSlots = 0;
    }

    @Override
    public void func_145839_a(NBTTagCompound tag) {
        super.func_145839_a(tag);

        if (tag.func_74764_b("EjectCounter"))
            this.ejectCounter = tag.func_74762_e("EjectCounter");
        if (tag.func_74764_b("XPCount"))
            this.experienceCount = tag.func_74762_e("XPCount");
        if (tag.func_74764_b("FilterType"))
            this.filterType = tag.func_74765_d("FilterType");
        if (tag.func_74764_b("Souls"))
            this.soulsRetained = tag.func_74762_e("Souls");
        this.power = tag.func_74771_c("power");
        validateInventory();
    }

    @Override
    public NBTTagCompound func_189515_b(NBTTagCompound tag) {
        NBTTagCompound t = super.func_189515_b(tag);
        t.func_74768_a("EjectCounter", this.ejectCounter);
        t.func_74768_a("XPCount", this.experienceCount);
        t.func_74777_a("FilterType", (short) this.filterType);
        t.func_74768_a("Souls", this.soulsRetained);
        t.func_74774_a("power", power);
        return t;
    }

    public boolean isPowered() {
        return power > 0;
    }

    private List<EntityItem> getCollidingItems(World world, BlockPos pos) {
        return world.func_175647_a(EntityItem.class, new AxisAlignedBB(pos.func_177958_n(), pos.func_177956_o() + 1, pos.func_177952_p(), pos.func_177958_n() + 1D, pos.func_177956_o() + 1.0001D, pos.func_177952_p() + 1D), EntitySelectors.field_94557_a);
    }

    private List<EntityXPOrb> getCollidingXPOrbs(World world, BlockPos pos) {
        return world.func_72872_a(EntityXPOrb.class, new AxisAlignedBB(pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p(), pos.func_177958_n() + 1D, pos.func_177956_o() + 1.5D, pos.func_177952_p() + 1D));
    }


    public boolean isXPFull() {
        return experienceCount >= maxExperienceCount;
    }

    private void insert() {
        if (!InvUtils.isFull(inventory)) {
            EntityItem item = getCollidingItems(field_145850_b, field_174879_c).stream().findFirst().orElse(null);
            if (item != null) {
                if (HopperInteractions.attemptToCraft(filterType, getBlockWorld(), getBlockPos(), item)) {
                    this.getBlockWorld().func_184148_a(null, field_174879_c.func_177958_n(), field_174879_c.func_177956_o(), field_174879_c.func_177952_p(), SoundEvents.field_187638_cR, SoundCategory.PLAYERS, 0.2F, ((getBlockWorld().field_73012_v.nextFloat() - getBlockWorld().field_73012_v.nextFloat()) * 0.7F + 1.0F) * 2.0F);
                }
                if (canFilterProcessItem(item.func_92059_d())) {
                    if (InvUtils.insertFromWorld(inventory, item, 0, 18, false))
                        this.getBlockWorld().func_184148_a(null, field_174879_c.func_177958_n(), field_174879_c.func_177956_o(), field_174879_c.func_177952_p(), SoundEvents.field_187638_cR, SoundCategory.PLAYERS, 0.2F, ((getBlockWorld().field_73012_v.nextFloat() - getBlockWorld().field_73012_v.nextFloat()) * 0.7F + 1.0F) * 2.0F);
                }
            }
        }

        if (!isXPFull() && filterType == 6) {
            List<EntityXPOrb> orbs = getCollidingXPOrbs(field_145850_b, field_174879_c);
            for (EntityXPOrb orb : orbs) {
                int remaining = maxExperienceCount - experienceCount;
                int value = orb.func_70526_d();
                if (remaining > 0) {
                    if (value <= remaining) {
                        this.experienceCount += value;
                        orb.func_70106_y();
                        continue;
                    }
                    orb.field_70530_e -= remaining;
                    this.experienceCount = maxExperienceCount;
                }
            }
        }
    }

    private void extract() {
        Optional<IItemHandler> inv = InvUtils.getItemHandler(field_145850_b, field_174879_c.func_177977_b(), EnumFacing.UP);
        if (ejectCounter > 2) {
            int slot = InvUtils.getFirstOccupiedStackInRange(inventory, 0, 17);
            if (slot != -1) {
                ItemStack stack = inventory.getStackInSlot(slot);
                if (inv.isPresent()) {
                    if (InvUtils.canInsert(inv.get(), stack, STACK_SIZE)) {
                        ItemStack insert = InvUtils.insert(inv.get(), stack, STACK_SIZE, false);
                        InvUtils.consumeItemsInInventory(inventory, stack, STACK_SIZE - insert.func_190916_E(), false);
                    }
                } else {
                    InvUtils.consumeItemsInInventory(inventory, stack, STACK_SIZE, false);
                    InvUtils.spawnStack(field_145850_b, field_174879_c.func_177958_n() + 0.5, field_174879_c.func_177956_o() - 0.5, field_174879_c.func_177952_p() + 0.5, STACK_SIZE, stack);
                }
            }
            ejectCounter = 0;
        } else {
            ejectCounter++;
        }
        if (ejectXPCounter > 2) {
            if (field_145850_b.func_180495_p(field_174879_c.func_177977_b()).func_185904_a().func_76222_j()) {
                if (experienceCount > 19) {
                    experienceCount -= 20;
                    spawnEntityXPOrb(20);
                }
            }
            ejectXPCounter = 0;
        } else {
            ejectXPCounter++;
        }
    }

    @Override
    public void func_73660_a() {
        byte power = (byte) calculateInput();
        if(this.power != power) {
            this.power = power;
        }
        if (!this.field_145850_b.field_72995_K) {
            insert();
            if (isPowered()) {
                extract();
            }
        }
        if (this.soulsRetained > 0) {
            processSouls();
        }
    }

    public boolean isUseableByPlayer(EntityPlayer player) {
        int x = field_174879_c.func_177958_n();
        int y = field_174879_c.func_177956_o();
        int z = field_174879_c.func_177952_p();
        return player.func_70092_e(x + 0.5D, y + 0.5D, z + 0.5D) <= 64.0D;
    }

    @Override
    public void func_70296_d() {
        super.func_70296_d();
        if (this.getBlockWorld() != null) {
            validateInventory();
        }
    }

    private boolean validateInventory() {
        boolean stateChanged = false;
        int currentFilter = getFilterType();
        if (currentFilter != this.filterType) {
            this.filterType = currentFilter;
            stateChanged = true;
        }
        byte slotsOccupied = (byte) InvUtils.getOccupiedStacks(inventory, 0, 17);
        if (slotsOccupied != this.occupiedSlots) {
            this.occupiedSlots = slotsOccupied;
            stateChanged = true;
        }
        if (getBlockWorld() != null && stateChanged) {
            IBlockState state = getBlockWorld().func_180495_p(field_174879_c);
            getBlockWorld().func_184138_a(field_174879_c, state, state, 3);
        }

        return stateChanged;
    }

    private int getFilterType() {
        ItemStack filter = getFilterStack();
        if (filter.func_190926_b()) {
            return -1;
        } else {
            return (short) HopperFilters.getFilterType(filter);
        }
    }

    private boolean canFilterProcessItem(ItemStack stack) {
        validateInventory();
        if (this.filterType == -1)
            return true;
        else if (this.filterType == 0)
            return this.getFilterStack().func_77969_a(stack);
        if (this.filterType > 0) {
            if (!this.isPowered())
                return false;
            if (HopperFilters.getAllowedItems(filterType) != null)
                return HopperFilters.getAllowedItems(filterType).test(stack);
        }
        return true;
    }

    private void spawnEntityXPOrb(int value) {
        double xOff = this.getBlockWorld().field_73012_v.nextDouble() * 0.1D + 0.45D;
        double yOff = -0.5D;
        double zOff = this.getBlockWorld().field_73012_v.nextDouble() * 0.1D + 0.45D;
        EntityXPOrb orb = new EntityXPOrb(this.getBlockWorld(), this.field_174879_c.func_177958_n() + xOff, this.field_174879_c.func_177956_o() + yOff, this.field_174879_c.func_177952_p() + zOff, value);
        orb.field_70159_w = 0.0D;
        orb.field_70181_x = 0.0D;
        orb.field_70179_y = 0.0D;
        this.getBlockWorld().func_72838_d(orb);
    }

    private void processSouls() {
        BlockPos down = field_174879_c.func_177977_b();
        if (this.filterType == 6) {
            Block blockBelow = this.getBlockWorld().func_180495_p(down).func_177230_c();
            if (soulsRetained > 0 && blockBelow instanceof ISoulSensitive && ((ISoulSensitive) blockBelow).isSoulSensitive(getBlockWorld(), down)) {
                int soulsConsumed = ((ISoulSensitive) blockBelow).processSouls(this.getBlockWorld(), down, this.soulsRetained);
                if (((ISoulSensitive) blockBelow).consumeSouls(this.getBlockWorld(), down, soulsConsumed))
                    this.soulsRetained -= soulsConsumed;
            } else if (soulsRetained > 7 && !isPowered()) {
                if (WorldUtils.spawnGhast(field_145850_b, field_174879_c))
                    this.getBlockWorld().func_184133_a(null, this.field_174879_c, SoundEvents.field_189105_bM, SoundCategory.BLOCKS, 1.0F, getBlockWorld().field_73012_v.nextFloat() * 0.1F + 0.8F);
                overpower();
            }
        } else {
            this.soulsRetained = 0;
        }
    }

    public void increaseSoulCount(int numSouls) {
        this.soulsRetained += numSouls;
    }

    @Override
    public int getInventorySize() {
        return 19;
    }

    @Override
    public SimpleStackHandler createItemStackHandler() {
        return new HopperHandler(getInventorySize(), this);
    }

    @Override
    public String getName() {
        return "inv.filtered_hopper.name";
    }

    @Override
    public int getSubtype() {
        return this.filterType;
    }

    @Override
    public void setSubtype(int type) {
        this.filterType = (short) Math.min(type, 7);
    }

    @Override
    public int getMaxVisibleSlots() {
        return 18;
    }

    public ModelWithResource getModel() {
        return RenderUtils.getModelFromStack(getFilterStack());
    }

    public ItemStack getFilterStack() {
        return inventory.getStackInSlot(18);
    }

    private class HopperHandler extends SimpleStackHandler {
        TileEntityFilteredHopper hopper;

        public HopperHandler(int size, TileEntityFilteredHopper hopper) {
            super(size, hopper);
            this.hopper = hopper;
        }

        @Override
        public void onContentsChanged(int slot) {
            super.onContentsChanged(slot);
            field_145850_b.func_175704_b(field_174879_c, field_174879_c);
            getBlockWorld().func_184138_a(field_174879_c, field_145850_b.func_180495_p(field_174879_c), field_145850_b.func_180495_p(field_174879_c), 2);
        }

        @Override
        public int getSlotLimit(int slot) {
            return slot == 18 ? 1 : super.getSlotLimit(slot);
        }
    }

    @Override
    public int getMechanicalOutput(EnumFacing facing) {
        return -1;
    }

    @Override
    public int getMechanicalInput(EnumFacing facing) {
        if (facing.func_176740_k().func_176722_c())
            return MechanicalUtil.getPowerOutput(field_145850_b, field_174879_c.func_177972_a(facing), facing.func_176734_d());
        return 0;
    }

    @Override
    public int getMaximumInput(EnumFacing facing) {
        return 1;
    }

    @Override
    public int getMinimumInput(EnumFacing facing) {
        return 0;
    }

    @Override
    public boolean hasCapability(@Nonnull Capability<?> capability, @Nonnull EnumFacing facing) {
        if (capability == CapabilityMechanicalPower.MECHANICAL_POWER)
            return true;
        return super.hasCapability(capability, facing);
    }

    @Nonnull
    @Override
    public <T> T getCapability(@Nonnull Capability<T> capability, @Nonnull EnumFacing facing) {
        if (capability == CapabilityMechanicalPower.MECHANICAL_POWER)
            return CapabilityMechanicalPower.MECHANICAL_POWER.cast(this);
        return super.getCapability(capability, facing);
    }

    @Override
    public World getBlockWorld() {
        return super.func_145831_w();
    }

    @Override
    public BlockPos getBlockPos() {
        return super.func_174877_v();
    }

    @Override
    public Block getBlock() {
        return func_145838_q();
    }
}
